/* db_sendfile.c - file send protocol
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <string.h>
#include "../common/db_packet.h"
#include "../runtime/db_vm.h"

/* prototypes */
static int WriteFile(FILE *fp, size_t fsize, char *dstname);
static int WriteSRAM(FILE *fp, size_t fsize);

int RunFile(char *name)
{
    int len = (int)strlen(name);

    /* make sure the name isn't too long */
    if (len >= PKTMAXLEN)
        return VMFALSE;

    /* send the packet */
    if (!PktSendPacket(PKT_RUN, (uint8_t *)name, len + 1))
		return VMFALSE;
    VM_printf("Program started\n");

	/* return successfully */
	return VMTRUE;
}

int SendFile(char *srcname, char *dstname)
{
    size_t fsize;
	FILE *fp;
    int sts;

    /* open the file */
    if (!(fp = fopen(srcname, "rb")))
        return VMFALSE;

    /* get the file size */
    fseek(fp, 0, SEEK_END);
    fsize = ftell(fp);
    fseek(fp, 0, SEEK_SET);

	/* send the file */
	if (dstname)
		sts = WriteFile(fp, fsize, dstname);
	else
		sts = WriteSRAM(fp, fsize);

    /* close the file */
    fclose(fp);

    /* return completion status */
    if (sts)
        VM_printf("Download complete            \n");
    else
        VM_printf("Download failed              \n");
    return sts;
}

static int WriteFile(FILE *fp, size_t fsize, char *dstname)
{
	int namelen = (int)strlen(dstname);
    uint8_t buf[PKTMAXLEN], *bp;
    int bc, ch, sts;

	/* make sure the name isn't too long */
	if (namelen >= PKTMAXLEN)
		return VMFALSE;

	/* send the filename header and the data */
	if (!PktSendPacket(PKT_OPEN, (uint8_t *)dstname, namelen + 1))
		return VMFALSE;

    /* start a sequence of packets */
    PktStartSequence();

    /* send each byte */
    for (bp = buf, bc = 0; (ch = getc(fp)) != EOF; ) {
        *bp++ = ch;
        if (++bc == sizeof(buf)) {
            printf("Remaining: %ld        \r", fsize);
            if (!(sts = PktSendPacket(PKT_DATA, buf, sizeof(buf))))
                return sts;
            fsize -= bc;
            bp = buf;
            bc = 0;
        }
    }

    /* flush partial buffer */
    if (bc > 0 && !PktSendPacket(PKT_DATA, buf, bc))
        return VMFALSE;

    /* send EOT and wait for ACK */
    return PktSendPacket(PKT_CLOSE, NULL, 0);
}

static int WriteSRAM(FILE *fp, size_t fsize)
{
	uint8_t buf[PKTMAXLEN];
    size_t remaining, cnt;
	FileHdr hdr;
	
	/* read the image header */
    if (fread(&hdr, 1, sizeof(hdr), fp) != sizeof(hdr))
		return VMFALSE;

	/* send the filename header and the data */
	if (!PktSendPacket(PKT_LOAD, (uint8_t *)&hdr, sizeof(hdr)))
		return VMFALSE;
	fsize -= sizeof(hdr);

    /* start a sequence of packets */
    PktStartSequence();

    /* read the code */
    fseek(fp, hdr.codeOffset, SEEK_SET);
	for (remaining = (size_t)hdr.codeSize; remaining > 0; remaining -= cnt) {
		if ((cnt = remaining) > PKTMAXLEN)
			cnt = PKTMAXLEN;
		if (fread(buf, 1, cnt, fp) != cnt)
			return VMFALSE;
        printf("Remaining: %ld        \r", (long)fsize);
		if (!PktSendPacket(PKT_LOADC, buf, (int)cnt))
			return VMFALSE;
		fsize -= (int)cnt;
	}

    /* read the initializers */
    fseek(fp, hdr.initOffset, SEEK_SET);
	for (remaining = (size_t)hdr.initSize; remaining > 0; remaining -= cnt) {
		if ((cnt = remaining) > PKTMAXLEN)
			cnt = PKTMAXLEN;
		if (fread(buf, 1, cnt, fp) != cnt)
			return VMFALSE;
        printf("Remaining: %ld        \r", (long)fsize);
		if (!PktSendPacket(PKT_LOADI, buf, (int)cnt))
			return VMFALSE;
		fsize -= (int)cnt;
	}

    /* start the downloaded program */
	return PktSendPacket(PKT_START, NULL, 0);
}
