/* db_bascom.h - definitions for a simple basic compiler
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#ifndef __DB_BASCOM_H__
#define __DB_BASCOM_H__

#include "setjmp.h"
#include "../runtime/db_vm.h"

#ifdef WIN32
#define strcasecmp  _stricmp
#endif

/* program limits */
#define MAXLINE     128
#define MAXTOKEN    32

/* globals */
extern jmp_buf errorTarget;

/* forward type declarations */
typedef struct SymbolTable SymbolTable;
typedef struct Symbol Symbol;
typedef struct TypeTable TypeTable;
typedef struct NamedType NamedType;
typedef struct Type Type;
typedef struct ParseTreeNode ParseTreeNode;
typedef struct ExprListEntry ExprListEntry;

/* lexical tokens */
typedef enum {
    T_NONE,
    T_REM = 0x100,  /* keywords start here */
    T_OPTION,
    T_DEF,
    T_SUB,
    T_TYPE,
    T_DIM,
    T_AS,
    T_LET,
    T_IF,
    T_THEN,
    T_ELSE,
    T_END,
    T_SELECT,
    T_CASE,
    T_FOR,
    T_TO,
    T_STEP,
    T_NEXT,
    T_DO,
    T_WHILE,
    T_UNTIL,
    T_LOOP,
    T_GOTO,
    T_CALL,
    T_MOD,
    T_AND,
    T_OR,
    T_NOT,
    T_STOP,
    T_PRINT,
    T_WRITE,
    T_DISPLAY,
    T_INPUT,
    T_OUTPUT,
    T_APPEND,
    T_OPEN,
    T_CLOSE,
    T_ELSE_IF,  /* compound keywords */
    T_END_DEF,
    T_END_SUB,
    T_END_IF,
    T_END_SELECT,
    T_DO_WHILE,
    T_DO_UNTIL,
    T_LOOP_WHILE,
    T_LOOP_UNTIL,
    T_LE,       /* non-keyword tokens */
    T_NE,
    T_GE,
    T_SHL,
    T_SHR,
    T_IDENTIFIER,
    T_NUMBER,
    T_STRING,
    T_EOL,
    T_EOF
} Token;

typedef enum {
    BLOCK_NONE,
    BLOCK_IF,
    BLOCK_ELSE,
    BLOCK_SELECT,
    BLOCK_CASE_ELSE,
    BLOCK_FOR,
    BLOCK_DO
} BlockType;

typedef struct Block Block;
struct Block {
    BlockType type;
    union {
        struct {
            int nxt;
            int end;
        } IfBlock;
        struct {
            int end;
        } ElseBlock;
        struct {
            int first;
            int nxt;
            int end;
        } SelectBlock;
        struct {
            int end;
        } CaseElseBlock;
        struct {
            int nxt;
            int end;
        } ForBlock;
        struct {
            int nxt;
            int end;
        } DoBlock;
    };
};

typedef struct String String;
struct String {
    String *next;
    int placed;
    int offset;
    int fixups;
    char value[1];
};

typedef struct Label Label;
struct Label {
    Label *next;
    int offset;
    int fixups;
    char name[1];
};

/* storage class ids */
typedef enum {
    SC_CONSTANT,
    SC_VARIABLE
} StorageClass;

/* symbol table */
struct SymbolTable {
    Symbol *head;
    Symbol **pTail;
};

/* symbol structure */
struct Symbol {
    Symbol *next;
    StorageClass storageClass;
    Type *type;
    int offset;
    int16_t value;
    char name[1];
};

/* type table */
struct TypeTable {
    NamedType *head;
    NamedType **pTail;
};

/* named type structure */
struct NamedType {
    NamedType *next;
    Type *type;
    char name[1];
};

/* type ids */
typedef enum {
    TYPE_UNKNOWN,
    TYPE_INTEGER,
    TYPE_BYTE,
    TYPE_STRING,
    TYPE_ARRAY,
    TYPE_FUNCTION
} TypeID;

/* type definition */
struct Type {
    TypeID  id;
    union {
        struct {
            int length;
        } StringInfo;
        struct {
            Type *elementType;
            int16_t dim1;
            int16_t dim2;
        } ArrayInfo;
        struct {
            Type *returnType;
            SymbolTable arguments;
            int argumentCount;
        } FunctionInfo;
    };
};

/* parse file */
typedef struct ParseFile ParseFile;
struct ParseFile {
    ParseFile *next;        /* next file in stack */
    FILE *fp;               /* input file pointer */
    char lineBuf[MAXLINE];  /* current input line */
    char *linePtr;          /* pointer to the current character */
    int lineNumber;         /* current line number */
    char name[1];           /* file name */
};

/* parse context */
typedef struct ParseContext ParseContext;
struct ParseContext {
    ParseFile *currentFile; /* scan - current file */
    Token savedToken;       /* scan - lookahead token */
    int tokenOffset;        /* scan - offset to the start of the current token */
    char token[MAXTOKEN];   /* scan - current token string */
    int value;              /* scan - current token integer value */
    int inComment;          /* scan - inside of a slash/star comment */
    Type integerType;       /* parse - integer type */
    Type integerArrayType;  /* parse - integer array type */
    Type byteType;          /* parse - byte type */
    Type byteArrayType;     /* parse - byte array type */
    Type stringType;        /* parse - string type */
    TypeTable types;        /* parse - type table */
    SymbolTable globals;    /* parse - global variables and constants */
    int globalCount;        /* parse - global variable count */
    String *strings;        /* parse - string constants */
    Label *labels;          /* parse - local labels */
    Symbol *function;       /* parse - symbol for current function definition */
    Type *functionType;     /* parse - type of the current function */
    SymbolTable locals;     /* parse - local variables of current function definition */
    int localCount;         /* parse - local variable count */
    Block blockBuf[10];     /* parse - stack of nested blocks */
    Block *bptr;            /* parse - current block */
    Block *btop;            /* parse - top of block stack */
    uint16_t codeOff;       /* bytecode - offset to next available code byte */
    uint8_t codeBuf[1024];  /* bytecode - code buffer */
    uint8_t *cptr;          /* bytecode - next available code position */
    uint8_t *ctop;          /* bytecode - top of code buffer */
    FileHdr hdr;            /* output - file header */
    FILE *ofp;              /* output - output file pointer */
    int debugMode;			/* output - VMTRUE to enable debug output */
    int heapSizeSet;        /* output - VMTRUE if the string heap size has been set */
};

/* partial value function codes */
typedef enum {
    PV_LOAD,
    PV_STORE
} PValOp;

/* partial value structure */
typedef struct PVAL PVAL;
struct PVAL {
    Type *type;
    void (*fcn)(ParseContext *c, PValOp op, PVAL *pv);
    union {
        String *str;
        int val;
    };
};

/* parse tree node types */
enum {
    NodeTypeSymbolRef,
    NodeTypeStringLit,
    NodeTypeIntegerLit,
    NodeTypeFunctionLit,
    NodeTypeUnaryOp,
    NodeTypeBinaryOp,
    NodeTypeArrayRef,
    NodeTypeFunctionCall,
    NodeTypeDisjunction,
    NodeTypeConjunction
};

/* parse tree node structure */
struct ParseTreeNode {
    int nodeType;
    union {
        struct {
            Symbol *symbol;
            void (*fcn)(ParseContext *c, PValOp op, PVAL *pv);
            int offset;
        } symbolRef;
        struct {
            String *string;
        } stringLit;
        struct {
            int16_t value;
        } integerLit;
        struct {
            Type *type;
            int offset;
        } functionLit;
        struct {
            int op;
            ParseTreeNode *expr;
            Type *resultType;
        } unaryOp;
        struct {
            int op;
            ParseTreeNode *left;
            ParseTreeNode *right;
            Type *resultType;
        } binaryOp;
        struct {
            ParseTreeNode *array;
            ParseTreeNode *index1;
            ParseTreeNode *index2;
        } arrayRef;
        struct {
            ParseTreeNode *fcn;
            ExprListEntry *args;
            int argc;
        } functionCall;
        struct {
            ExprListEntry *exprs;
        } exprList;
    };
};

/* expression list entry structure */
struct ExprListEntry {
    ParseTreeNode *expr;
    ExprListEntry *next;
};

/* db_compiler.c */
int CompileFile(char *infile, char *outfile, int debugMode);
int DownloadImage(char *imagename, char *dstname);
void ConstructOutputName(char *infile, char *outfile);
int IsSourceFile(char *file);
int IsImageFile(char *file);
void SelectRuntime(ParseContext *c, char *target);
void SetHeapSize(ParseContext *c, char *sizes);
String *AddString(ParseContext *c, char *value);
int AddStringRef(String *str, int offset);
void WriteFunction(ParseContext *c, char *name);
void *Alloc(char *tag, size_t size);
void Free(void *p);
void FreeAll(void);
void DumpAll(void);
void Fatal(char *fmt, ...);

/* db_statement.c */
void ParseStatement(ParseContext *c);

/* db_expr.c */
void ParseRValue(ParseContext *c);
ParseTreeNode *ParseExpr(ParseContext *c);
ParseTreeNode *ParseSubroutineCall(ParseContext *c, ParseTreeNode *functionNode);
ParseTreeNode *ParsePossibleArrayReference(ParseContext *c, ParseTreeNode *node);
ParseTreeNode *GetSymbolRef(ParseContext *c, char *name);
void FreeParseTree(ParseTreeNode *expr);
Type *NodeSrcType(ParseContext *c, ParseTreeNode *expr);
Type *NodeType(ParseContext *c, ParseTreeNode *expr);
int IsIntegerLit(ParseTreeNode *node);

/* db_scan.c */
void InitParseContext(ParseContext *c);
int PushFile(ParseContext *c, char *name);
void CloseParseContext(ParseContext *c);
int GetLine(ParseContext *c);
void FRequire(ParseContext *c, Token requiredToken);
void Require(ParseContext *c, Token token, Token requiredToken);
int GetToken(ParseContext *c);
void SaveToken(ParseContext *c, Token token);
char *TokenName(Token token);
int SkipSpaces(ParseContext *c);
int GetChar(ParseContext *c);
void UngetC(ParseContext *c);
void ParseError(ParseContext *c, char *fmt, ...);

/* db_symbols.c */
void InitSymbolTable(SymbolTable *table);
void FreeSymbols(ParseContext *c, SymbolTable *table);
Symbol *AddSymbol(SymbolTable *table, char *name, StorageClass storageClass, Type *type, int offset, int value);
Symbol *FindSymbol(SymbolTable *table, char *name);
void DumpSymbols(SymbolTable *table, int showTags);

/* db_type.c */
void InitTypeTable(TypeTable *table);
void FreeTypes(ParseContext *c, TypeTable *table);
Type *NewType(int id);
void FreeType(ParseContext *c, Type *type);
Type *DefaultType(ParseContext *c, char *name);
Type *AddType(TypeTable *table, char *name, Type *type);
Type *FindType(TypeTable *table, char *name);
int CompareTypes(Type *type1, Type *type2);
int ValueSize(Type *type, int *pPrefix);
int ValueByteSize(Type *type);

/* db_bytecode.c */
void code_lvalue(ParseContext *c, ParseTreeNode *expr, PVAL *pv);
void code_rvalue(ParseContext *c, ParseTreeNode *expr);
void rvalue(ParseContext *c, PVAL *pv);
void chklvalue(ParseContext *c, PVAL *pv);
void code_literal(ParseContext *c, PValOp fcn, PVAL *pv);
void code_global(ParseContext *c, PValOp fcn, PVAL *pv);
void code_argument(ParseContext *c, PValOp fcn, PVAL *pv);
void code_local(ParseContext *c, PValOp fcn, PVAL *pv);
void code_index(ParseContext *c, PValOp fcn, PVAL *pv);
int codeaddr(ParseContext *c);
int putcbyte(ParseContext *c, int b);
int putcword(ParseContext *c, int w);
int merge(ParseContext *c, int chn, int chn2);
void fixup(ParseContext *c, int chn, int val);
void fixupbranch(ParseContext *c, int chn, int val);

#endif

