/* db_vm.h - definitions for a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#ifndef __DB_VM_H__
#define __DB_VM_H__

#include <stdio.h>
#include <stdarg.h>
#include <setjmp.h>
#include "../common/db_types.h"

/* forward type declarations */
typedef struct Interpreter Interpreter;

/* image file signature */
#define FILE_TAG        "DMBASIC"
#define FILE_VERSION    0x0105

/* target runtimes */
#define TARGET_UNKNOWN      0
#define TARGET_BITMAP       1
#define TARGET_TILE         2

/* image file header */
typedef struct {
    char        tag[8];         /* must be 'DMBASIC' */
    uint16_t    version;        /* must be equal to FILE_VERSION above */
    uint8_t     target;         /* target runtime */
    uint8_t     unused;         /* (unused) */
    uint16_t    codeOffset;     /* offset to the bytecodes */
    uint16_t    codeSize;       /* size of bytecodes in bytes */
    uint16_t    mainCodeOffset; /* offset to the main code */
    uint16_t    initOffset;     /* offset to the initialization table */
    uint16_t    initSize;       /* size of the initialization table in bytes */
    uint16_t    variableCount;  /* number of variables */
    uint16_t    stackSize;      /* stack size */
    uint16_t    heapSize;       /* string heap size */
    uint16_t    heapHandles;    /* number of heap handles */
} FileHdr;

/* intrinsic function handler type */
typedef void IntrinsicFcn(Interpreter *i);

/* string handle structure */
typedef struct {
    uint8_t *data;
} StringHandle;

/* string header structure */
typedef struct {
    StringHandle *handle;
    uint16_t refCnt;
    uint16_t length;
} StringHdr;

/* macro to get a pointer to an object in the string heap */
#define GetHeapStringPtr(i, h)  (((i)->heap->endHandles - (h))->data)
#define GetHeapStringLen(i, h)  (((StringHdr *)GetHeapStringPtr(i, h) - 1)->length)

/* string heap structure */
typedef struct {
    int nHandles;               /* number of handles */
    StringHandle *handles;      /* array of handles */
    StringHandle *endHandles;   /* end of the array of handles */
    StringHandle *freeHandles;  /* list of free handles */
    uint8_t *data;              /* heap data */
    uint8_t *free;              /* next free heap location */
} StringHeap;

/* interpreter states */
#define STATE_IDLE      0
#define STATE_LOADING   1
#define STATE_STARTING  2
#define STATE_RUNNING   3
#define STATE_PAUSED    4

/* interpreter state structure */
struct Interpreter {
    FileHdr hdr;
    jmp_buf errorTarget;
    int init_state;
    uint16_t init_off;
    uint16_t init_cnt;
    int16_t *variables;
    uint8_t *code;
    int16_t *stack;
    int16_t *stackTop;
    int16_t *fp;
    int16_t *sp;
    uint8_t *pc;
    StringHeap *heap;
    uint16_t linePos;
    int state;
    int cycleCount;
    int target;
};

/* initializer processing states */
#define INIT_HDR1   1
#define INIT_HDR2   2
#define INIT_DATA   3

/* stack manipulation macros */
#define Reserve(i, n)   do {                                    \
                            if ((i)->sp + (n) > (i)->stackTop)  \
                                StackOverflow(i);               \
                            else  {                             \
                                int cnt = (n);                  \
                                while (--cnt >= 0)              \
                                    Push(i, 0);                 \
                            }                                   \
                        } while (0)
#define CPush(i, v)     do {                                    \
                            if ((i)->sp + 1 >= (i)->stackTop)   \
                                StackOverflow(i);               \
                            else                                \
                                Push(i, v);                     \
                        } while (0)
#define Push(i, v)      (*++(i)->sp = (v))
#define Pop(i)          (*(i)->sp--)
#define Drop(i, n)      ((i)->sp -= (n))

/* prototypes */
void StackOverflow(Interpreter *i);

/* opcodes */
#define OP_HALT         0x00    /* halt */
#define OP_BRT          0x01    /* branch on true */
#define OP_BRTSC        0x02    /* branch on true (for short circuit booleans) */
#define OP_BRF          0x03    /* branch on false */
#define OP_BRFSC        0x04    /* branch on false (for short circuit booleans) */
#define OP_BR           0x05    /* branch unconditionally */
#define OP_NOT          0x06    /* logical negate top of stack */
#define OP_NEG          0x07    /* negate */
#define OP_ADD          0x08    /* add two numeric expressions */
#define OP_SUB          0x09    /* subtract two numeric expressions */
#define OP_MUL          0x0a    /* multiply two numeric expressions */
#define OP_DIV          0x0b    /* divide two numeric expressions */
#define OP_REM          0x0c    /* remainder of two numeric expressions */
#define OP_BNOT         0x0d    /* bitwise not of two numeric expressions */
#define OP_BAND         0x0e    /* bitwise and of two numeric expressions */
#define OP_BOR          0x0f    /* bitwise or of two numeric expressions */
#define OP_BXOR         0x10    /* bitwise exclusive or */
#define OP_SHL          0x11    /* shift left */
#define OP_SHR          0x12    /* shift right */
#define OP_LT           0x13    /* less than */
#define OP_LE           0x14    /* less than or equal to */
#define OP_EQ           0x15    /* equal to */
#define OP_NE           0x16    /* not equal to */
#define OP_GE           0x17    /* greater than or equal to */
#define OP_GT           0x18    /* greater than */
#define OP_CAT          0x19    /* concatenate two strings */
#define OP_LTS          0x1a    /* string less than */
#define OP_LES          0x1b    /* string less than or equal to */
#define OP_EQS          0x1c    /* string equal to */
#define OP_NES          0x1d    /* string not equal to */
#define OP_GES          0x1e    /* string greater than or equal to */
#define OP_GTS          0x1f    /* string greater than */
#define OP_LIT          0x20    /* load literal */
#define OP_GREF         0x21    /* load a global variable */
#define OP_GSET         0x22    /* set a global variable */
#define OP_GREFB        0x23    /* load a global byte variable */
#define OP_GSETB        0x24    /* set a global byte variable */
#define OP_GREFS        0x25    /* load a global string variable */
#define OP_GSETS        0x26    /* set a global string variable */
#define OP_LOFF         0x27    /* load the offset into variable space for a local variable */
#define OP_LREF         0x28    /* load a local variable relative to the frame pointer */
#define OP_LSET         0x29    /* set a local variable relative to the frame pointer */
#define OP_LREFB        0x2a    /* load a local byte variable relative to the frame pointer */
#define OP_LSETB        0x2b    /* set a local byte variable relative to the frame pointer */
#define OP_LREFS        0x2c    /* load a local string variable relative to the frame pointer */
#define OP_LSETS        0x2d    /* set a local string variable relative to the frame pointer */
#define OP_LDECREF      0x2e    /* decrement the reference count of a local variable */
#define OP_AREF1        0x2f    /* load an element of an array */
#define OP_ASET1        0x30    /* set an element of an array */
#define OP_AREF2        0x31    /* load an element of a two-dimensional array */
#define OP_ASET2        0x32    /* set an element of a two-dimensional array */
#define OP_AREF1B       0x33    /* load a byte element of an array */
#define OP_ASET1B       0x34    /* set a byte element of an array */
#define OP_AREF2B       0x35    /* load a byte element of a two-dimensional array */
#define OP_ASET2B       0x36    /* set a byte element of a two-dimensional array */
#define OP_AREF1S       0x37    /* load a string element of an array */
#define OP_ASET1S       0x38    /* set a string element of an array */
#define OP_AREF2S       0x39    /* load a string element of a two-dimensional array */
#define OP_ASET2S       0x3a    /* set a string element of a two-dimensional array */
#define OP_RESERVE      0x3b    /* reserve space on the stack */
#define OP_CALL         0x3c    /* call a function */
#define OP_FRET         0x3d    /* return from a function leaving an integer result on the stack */
#define OP_SRET         0x3e    /* return from a subroutine */
#define OP_DUP          0x3f    /* duplicate the top element of the stack */
#define OP_DROP         0x40    /* drop the top element of the stack */
#define OP_PRINTI       0x41    /* print an integer */
#define OP_PRINTS       0x42    /* print a string */
#define OP_PRINTT       0x43    /* print a tab */
#define OP_PRINTN       0x44    /* print a newline */
#define OP_GETLINE      0x45    /* get a new input line */
#define OP_INPUTI       0x46    /* input an integer */
#define OP_INPUTS       0x47    /* input a string */
#define OP_INPUTL       0x48    /* input a line */
#define OP_OPEN         0x49    /* open a file */
#define OP_CLOSE        0x4a    /* close a file */
#define OP_WRITES       0x4b    /* write a string */

/* stack frame offsets */
#define F_FP    1
#define F_PC    2
#define F_SIZE  2

/* intrinsic function numbers */
enum {
    FN_ABS              = 0x00,
    FN_RND,
    FN_LEFT,
    FN_RIGHT,
    FN_MID,
    FN_CHR,
    FN_STR,
    FN_VAL,
    FN_ASC,
    FN_LEN,
    FN_WAITFORVSYNC,
    FN_GAMEPAD,
    FN_SOUND,
    FN_PLOT,
    FN_LINE,
    FN_HLINE,
    FN_VLINE,
    FN_CLEAR,
    FN_FILLRECT,
    FN_CIRCLE,
    FN_POSITION,
    FN_TEXT,
    FN_SETTILEMAP,
    FN_SETTILES,
    FN_SETCOLORTABLE,
    FN_SETSPRITE
};

/* the only global used by the interpreter */
extern Interpreter interp;

/* prototypes and variables from db_vmfcn.c */
extern IntrinsicFcn * const Intrinsics[];

/* prototypes from db_vmimage.c  */
int LoadImageFile(Interpreter *i, char *name);

/* prototypes from db_vmint.c */
void InitInterpreter(Interpreter *i, int target);
int InitMemoryImage(Interpreter *i);
int StartImageLoad(Interpreter *i, FileHdr *hdr);
void ProcessInitData(Interpreter *i, uint8_t *buf, int len);
void FinishImageLoad(Interpreter *i);
int Execute(Interpreter *i, int cycleCount);
void SetInputLine(uint8_t *buf, int cnt);
void Abort(Interpreter *i, char *fmt, ...);

/* prototypes from db_vmdebug.c */
void DecodeFunction(uint16_t base, uint8_t *code, int len);
int DecodeInstruction(uint16_t base, uint8_t *code, uint8_t *lc);
void ShowStack(Interpreter *i);

/* prototypes from db_vmheap.c */
void InitHeap(Interpreter *i, uint16_t size, uint16_t nHandles);
int16_t StringAlloc(Interpreter *i, uint16_t length);
void StringAddRef(Interpreter *i, int16_t handle);
void StringRelease(Interpreter *i, int16_t handle);
char *GetStringPtr(Interpreter *i, int16_t value, int *pInHeap);
int GetStringLen(Interpreter *i, int16_t value);
void CompactHeap(Interpreter *i);
void DumpHeap(Interpreter *i);

/* prototypes for image memory allocation functions */
void VM_InitImage(void);
int VM_LoadMemoryImageHeader(Interpreter *i);
void VM_StoreMemoryImageHeader(Interpreter *i);
uint8_t *VM_GetMemoryImageInitData(Interpreter *i);
int VM_AllocateImage(Interpreter *i, uint16_t codeSize, uint16_t initSize, uint16_t dataSize, uint16_t heapSize);
uint8_t *VM_FreeSpace(int *pSize);
void VM_FreeImage(Interpreter *i);
void VM_LoadCode(uint16_t dstoff, uint8_t *src, int cnt);
void VM_LoadInit(uint16_t dstoff, uint8_t *src, int cnt);
void VM_ReadInit(uint16_t off, uint8_t *buf, int cnt);

/* functions provided by target platform */
void VM_sysinit(void);
void VM_fcninit(void);
void VM_fcnstop(void);
void VM_printf(char *fmt, ...);
void VM_vprintf(char *fmt, va_list ap);
void VM_putchar(int ch);
void VM_flush(void);
void VM_getline(void);
void VM_displayinit(void);
void VM_displaypos(int x, int y);
void VM_displaychar(int ch);
void VM_displaystr(char *str);
void VM_displaytab(void);
void VM_writedebug(char *buf, int cnt);
VMFILE *VM_fopen(const char *name, const char *mode);
int VM_fclose(VMFILE *file);
size_t VM_fread(void *buf, size_t count, VMFILE *file);
size_t VM_fwrite(const void *buf, size_t count, VMFILE *file);
int VM_fseek(VMFILE *file, long offset, int whence);
long VM_ftell(VMFILE *file);

/* platform specific built in functions */
void fcn_waitforvsync(Interpreter *i);
void fcn_gamepad(Interpreter *i);
void fcn_sound(Interpreter *i);
void fcn_plot(Interpreter *i);
void fcn_line(Interpreter *i);
void fcn_hline(Interpreter *i);
void fcn_vline(Interpreter *i);
void fcn_clear(Interpreter *i);
void fcn_fillrect(Interpreter *i);
void fcn_circle(Interpreter *i);
void fcn_position(Interpreter *i);
void fcn_text(Interpreter *i);
void fcn_settilemap(Interpreter *i);
void fcn_settiles(Interpreter *i);
void fcn_setcolortable(Interpreter *i);
void fcn_setsprite(Interpreter *i);

#endif
