/* db_vmfcn.c - intrinsic functions
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdlib.h>
#include <ctype.h>
#include "db_vm.h"

static void fcn_abs(Interpreter *i);
static void fcn_rnd(Interpreter *i);
static void fcn_left(Interpreter *i);
static void fcn_right(Interpreter *i);
static void fcn_mid(Interpreter *i);
static void fcn_chr(Interpreter *i);
static void fcn_str(Interpreter *i);
static void fcn_val(Interpreter *i);
static void fcn_asc(Interpreter *i);
static void fcn_len(Interpreter *i);

/* this table must be in the same order as the FN_xxx macros */
IntrinsicFcn * const Intrinsics[] = {
    fcn_abs,
	fcn_rnd,
    fcn_left,
    fcn_right,
    fcn_mid,
    fcn_chr,
    fcn_str,
    fcn_val,
    fcn_asc,
    fcn_len,
	fcn_waitforvsync,
	fcn_gamepad,
	fcn_sound,
	fcn_plot,
	fcn_line,
    fcn_hline,
    fcn_vline,
	fcn_clear,
	fcn_fillrect,
	fcn_circle,
    fcn_position,
    fcn_text,
    fcn_settilemap,
    fcn_settiles,
    fcn_setcolortable,
    fcn_setsprite
};

/* local function prototypes */
static int GetStringVal(char *str, int len, int inHeap);

/* fcn_abs - ABS(n): return the absolute value of a number */
static void fcn_abs(Interpreter *i)
{
    int n = i->sp[0];
    i->sp[-1] = abs(n);
    Drop(i, 1);
}

/* fcn_rnd - RND(n): return a random number between 0 and n-1 */
static void fcn_rnd(Interpreter *i)
{
    int n = i->sp[0];
    i->sp[-1] = rand() % n;
    Drop(i, 1);
}

/* fcn_left - LEFT$(str$, n): return the leftmost n characters of a string */
static void fcn_left(Interpreter *i)
{
	int inHeap;
    char *str = GetStringPtr(i, i->sp[-1], &inHeap);
    int len = GetStringLen(i, i->sp[-1]);
    int16_t n = i->sp[0];
    if (n > len)
        n = len;
    i->sp[-2] = StringAlloc(i, n);
    VMCOPY(GetHeapStringPtr(i, i->sp[-2]), str, n, inHeap);
    StringRelease(i, i->sp[-1]);
    Drop(i, 2);
}

/* fcn_right - RIGHT$(str$, n): return the rightmost n characters of a string */
static void fcn_right(Interpreter *i)
{
	int inHeap;
    char *str = GetStringPtr(i, i->sp[-1], &inHeap);
    int len = GetStringLen(i, i->sp[-1]);
    int16_t n = i->sp[0];
    if (n > len)
        n = len;
    i->sp[-2] = StringAlloc(i, n);
    VMCOPY(GetHeapStringPtr(i, i->sp[-2]), str + len - n, n, inHeap);
    StringRelease(i, i->sp[-1]);
    Drop(i, 2);
}

/* fcn_mid - MID$(str$, start, n): return n characters from the middle of a string */
static void fcn_mid(Interpreter *i)
{
	int inHeap;
    char *str = GetStringPtr(i, i->sp[-2], &inHeap);
    int len = GetStringLen(i, i->sp[-2]);
    int16_t start = i->sp[-1] - 1;
    int16_t n = i->sp[0];
    if (start < 0 || start >= len)
        Abort(i, "string index out of range: %d", start + 1);
    if (start + n > len)
        n = len - start;
    i->sp[-3] = StringAlloc(i, n);
    VMCOPY(GetHeapStringPtr(i, i->sp[-3]), str + start, n, inHeap);
    StringRelease(i, i->sp[-2]);
    Drop(i, 3);
}

/* fcn_chr - CHR$(n): return a one character string with the specified character code */
static void fcn_chr(Interpreter *i)
{
    char buf[1];
    buf[0] = (char)i->sp[0];
    i->sp[-1] = StringAlloc(i, 1);
    VMCOPY(GetHeapStringPtr(i, i->sp[-1]), buf, 1, VMTRUE);
    Drop(i, 1);
}

/* fcn_str - STR$(n): return n converted to a string */
static void fcn_str(Interpreter *i)
{
    char buf[20];
    int len = sprintf(buf, "%d", i->sp[0]);
    i->sp[-1] = StringAlloc(i, (uint16_t)len);
    VMCOPY(GetHeapStringPtr(i, i->sp[-1]), buf, len, VMTRUE);
    Drop(i, 1);
}

/* fcn_val - VAL(str$): return the numeric value of a string */
static void fcn_val(Interpreter *i)
{
	int inHeap;
    char *str = GetStringPtr(i, i->sp[0], &inHeap);
    int len = GetStringLen(i, i->sp[0]);
    i->sp[-1] = GetStringVal(str, len, inHeap);
    Drop(i, 1);
}

/* fcn_asc - ASC(str$): return the character code of the first character of a string */
static void fcn_asc(Interpreter *i)
{
	int inHeap;
    char *str = GetStringPtr(i, i->sp[0], &inHeap);
    int len = GetStringLen(i, i->sp[0]);
    i->sp[-1] = len > 0 ? VMGETBYTE(str, inHeap) : 0;
    StringRelease(i, i->sp[0]);
    Drop(i, 1);
}

/* fcn_len - LEN(str$): return length of a string */
static void fcn_len(Interpreter *i)
{
    i->sp[-1] = GetStringLen(i, i->sp[0]);
    StringRelease(i, i->sp[0]);
    Drop(i, 1);
}

/* GetStringVal - get the numeric value of a string */
static int GetStringVal(char *str, int len, int inHeap)
{
    int radix = 'd';
    int sign = 1;
    int val = 0;
    int ch;

    /* handle non-empty strings */
    if (--len >= 0) {

        /* check for a sign or a radix indicator */
        switch (ch = VMGETBYTE(str++, inHeap)) {
        case '-':
            sign = -1;
            break;
        case '+':
            /* sign is 1 by default */
            break;
        case '0':
            if (--len < 0)
                return 0;
            switch (ch = VMGETBYTE(str++, inHeap)) {
            case 'b':
            case 'B':
                radix = 'b';
                break;
            case 'x':
            case 'X':
                radix = 'x';
                break;
            }
            break;
        default:
            if (!isdigit(ch))
                return 0;
            val = ch - '0';
            break;
        }


        /* get the number */
        switch (radix) {
        case 'b':
            while (--len >= 0) {
                ch = VMGETBYTE(str++, inHeap);
                if (ch == '0' || ch == '1')
                    val = val * 2 + ch - '0';
                else if (ch != '_')
                    break;
            }
            break;
        case 'd':
            while (--len >= 0) {
                ch = VMGETBYTE(str++, inHeap);
                if (isdigit(ch))
                    val = val * 10 + ch - '0';
                else if (ch != '_')
                    break;
            }
            break;
        case 'x':
            while (--len >= 0) {
                ch = VMGETBYTE(str++, inHeap);
                if (isxdigit(ch)) {
                    int lowerch = tolower(ch);
                    val = val * 16 + (isdigit(lowerch) ? lowerch - '0' : lowerch - 'a' + 16);
                }
                else if (ch != '_')
                    break;
            }
            break;
        }
    }
    
    /* return the numeric value */
    return sign * val;
}
