/* db_main.c - basic compiler command line program
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <setjmp.h>
#include <string.h>
#include <ctype.h>
#include "../compiler/db_compiler.h"
#include "../common/db_packet.h"

/* error target */
static jmp_buf errorTarget;

/* local function prototypes */
static Usage(void);
static void DownloadFile(char *port, char *file, char *dstfile);

/* main - the main routine */
int main(int argc, char *argv[])
{
    int compileOnly, terminalMode, debugMode, i;
    char *file, *dstfile, *port;

    /* initialize */
    file = NULL;
    dstfile = NULL;
    port = "COM1";
    compileOnly = VMFALSE;
    terminalMode = VMFALSE;
    debugMode = VMFALSE;

    /* get the arguments */
    for(i = 1; i < argc; ++i) {

        /* handle switches */
        if(argv[i][0] == '-') {
            switch(argv[i][1]) {
            case 'c':
                dstfile = "autorun.bai";
                break;
            case 'd':
                debugMode = VMTRUE;
                break;
            case 'f':
                dstfile = NULL;
                break;
            case 'p':
                if(argv[i][2])
                    port = &argv[i][2];
                else if(++i < argc)
                    port = argv[i];
                else
                    Usage();
                break;
            case 't':
                terminalMode = VMTRUE;
                break;
            case 'x':
                compileOnly = VMTRUE;
                break;
            default:
                Usage();
                break;
            }
        }

        /* handle the input filename */
        else {
            if (file)
                Usage();
            file = argv[i];
        }
    }

    /* check for a numeric port */
    if (isdigit(port[0])) {
        static char buf[10];
        sprintf(buf, "COM%d", atoi(port));
        port = buf;
    }

    /* handle direct entry into terminal mode */
    if (terminalMode)
        DownloadFile(port, NULL, NULL);

    /* compile and download or just download a file */
    else {
        
        /* make sure there was at least an input file specified */
        if (!file)
            Usage();

        /* ".bas" is a source file */
        if (IsSourceFile(file)) {
            char outfile[_MAX_PATH];
            ConstructOutputName(file, outfile);
            if (CompileFile(file, outfile, debugMode) && !compileOnly)
                DownloadFile(port, outfile, dstfile);
        }

        /* ".bai" is an already compiled image file */
        else if (IsImageFile(file))
            DownloadFile(port, file, dstfile);

        /* unknown file extension */
        else
            Usage();
    }

    /* exit successfully */
    return 0;
}

/* Usage - display a usage message and exit */
static Usage(void)
{
    fprintf(stderr, "\
usage: xgsbasic\n\
         [ -p <port> ]     serial port (default is COM1)\n\
         [ -x ]            compile only and don't download\n\
         [ -c ]            downloads to the microSD card (autorun.bai)\n\
         [ -f ]            downloads to the on-chip flash (default)\n\
         [ -t ]            enter terminal mode with no compile/download\n\
         [ <name>.<ext> ]  <source>.bas or <compiled>.bai\n\
\n\
   for example:\n\
\n\
     xgsbasic -p COM6 pong.bas\n\
\n\
   will download the file 'pong.bas' to on-chip flash using serial port COM6\n\
\n\
   and:\n\
\n\
     xgsbasic -c life.bas\n\
\n\
   will download the file 'life.bas' to the microSD card\n\
\n\
     xgsbasic life.bai\n\
\n\
   will download the already compiled file to on-chip flash\n\
\n\
   After the download is complete, terminal mode is entered to display\n\
   any output produced by the downloaded program.\n\
\n\
   If -t is given, no compile or download is performed and terminal mode\n\
   is entered directly.\n\
");
    exit(1);
}

/* DownloadFile - download a file to the target system */
static void DownloadFile(char *port, char *file, char *dstfile)
{
    /* open the communications line */
    if (!PktOpen(port)) {
        fprintf(stderr, "error: can't open serial port: %s\n", port);
        exit(1);
    }

    /* download the image */
    if (file)
        DownloadImage(file, dstfile);

    /* handle requests from the runtime */
	for (;;)
        PktPoll(0);

    /* close the communications line */
    PktClose();
}

void VM_writedebug(char *buf, int cnt)
{
    fwrite(buf, cnt, 1, stderr);
}

void VM_printf(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);
}

void VM_vprintf(char *fmt, va_list ap)
{
	vprintf(fmt, ap);
}

void VM_putchar(int ch)
{
	putchar(ch);
}

void VM_getline(void)
{
    char buf[PKTMAXLEN];
    while (!fgets(buf, sizeof(buf), stdin))
        ;
    PktSendFrame(PKT_LINE, buf, strlen(buf));
}

