/* listener.cpp - an interactive text window
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdarg.h>

#include "listener.h"

BEGIN_EVENT_TABLE(Listener,wxTextCtrl)
    EVT_CHAR(Listener::OnChar)
END_EVENT_TABLE()

Listener::Listener(
                wxWindow *parent,
                wxWindowID id,
                const wxString &value,
                const wxPoint &pos,
                const wxSize &size,
                int style)
    : wxTextCtrl(parent,id,value,pos,size,style | wxTE_RICH | wxTE_MULTILINE | wxTE_DONTWRAP | wxHSCROLL | wxVSCROLL),
      m_font(9,wxFONTFAMILY_MODERN,wxFONTSTYLE_NORMAL,wxFONTWEIGHT_NORMAL),
      m_inStyle(*wxBLUE,wxNullColour,m_font),
      m_outStyle(*wxBLACK,wxNullColour,m_font),
      m_inputHandler(NULL)
{
    SetDefaultStyle(m_inStyle);
}

void Listener::Printf(wxString fmt,...)
{
    va_list ap;
    va_start(ap,fmt);
    m_outBuf.Printf(fmt,ap);
    if (CheckForEOL())
        FlushOutput();
    va_end(ap);
}

void Listener::PutS(wxString str)
{
    m_outBuf.Append(str);
    if (CheckForEOL())
        FlushOutput();
}

void Listener::PutC(wxChar ch)
{
    m_outBuf.Append(ch);
    if (ch == '\n')
        FlushOutput();
}

void Listener::FlushOutput()
{
    Append(m_outBuf);
    m_outBuf.Empty();
}

bool Listener::CheckForEOL()
{
    return m_outBuf.Find('\n') != wxNOT_FOUND;
}

void Listener::Append(wxString str)
{
    SetDefaultStyle(m_outStyle);
    AppendText(str);
    SetDefaultStyle(m_inStyle);

    wxTextPos last = GetLastPosition();
    SetSelection(last,last);
}

void Listener::SetInputStyle()
{
    SetDefaultStyle(m_inStyle);
}

void Listener::SetOutputStyle()
{
    SetDefaultStyle(m_outStyle);
}

void Listener::SetInputHandler(void (*handler)(void *data, const char *buf), void *data)
{
    m_inputHandler = handler;
    m_inputHandlerData = data;
}

void Listener::OnChar(wxKeyEvent &event)
{
    switch (event.GetKeyCode()) {
    case WXK_RETURN:
        if (event.ControlDown())
            event.Skip();
        else
            HandleUserInput();
        break;
    default:
        event.Skip();
        break;
    }
}

void Listener::HandleUserInput()
{
    if (m_inputHandler) {
        long start, end;
        GetSelection(&start, &end);

        if (start == end)
            FindInputText(start, &start, &end);
    
        if (end >= start) {

            SetSelection(start, end);

            wxString str = GetStringSelection();
            str.Append('\n');

            SetSelection(end, end);
            SetOutputStyle();
            WriteText(wxT("\n"));
            SetInputStyle();

            (*m_inputHandler)(m_inputHandlerData, (char *)(const wxChar *)str);
        }
    }
}

void Listener::FindInputText(long current, long *pStart, long *pEnd)
{
    long x, y;
    PositionToXY(current, &x, &y);

    long length = GetLineLength(y);
    long end = XYToPosition(0, y) + length;

    while (current > 0) {
        wxTextAttr style;
        GetStyle(current - 1,style);
        if (style.GetTextColour() != m_inStyle.GetTextColour())
            break;
        --current;
    }

    *pStart = current;
    *pEnd = end;
}
